@@ -1,6 +1,6 @@
-#Copyright ReportLab Europe Ltd. 2000-2016
+#Copyright ReportLab Europe Ltd. 2000-2017
 #see license.txt for license details
-#history http://www.reportlab.co.uk/cgi-bin/viewcvs.cgi/public/reportlab/trunk/reportlab/graphics/shapes.py
+#history https://hg.reportlab.com/hg-public/reportlab/log/tip/src/reportlab/graphics/shapes.py
 
 __version__='3.3.0'
 __doc__='''Core of the graphics library - defines Drawing and Shapes'''
@@ -20,6 +20,8 @@ from reportlab.lib.attrmap import *
 from reportlab.lib.rl_accel import fp_str
 from reportlab.pdfbase.pdfmetrics import stringWidth
 from reportlab.lib.fonts import tt2ps
+from reportlab.pdfgen.canvas import FILL_EVEN_ODD, FILL_NON_ZERO
+from reportlab import xrange
 _baseGFontNameB = tt2ps(_baseGFontName,1,0)
 _baseGFontNameI = tt2ps(_baseGFontName,0,1)
 _baseGFontNameBI = tt2ps(_baseGFontName,1,1)
@@ -55,14 +57,13 @@ STATE_DEFAULTS = {   # sensible defaults for all
     'overprintMask': 0,
 
     'fillColor': colors.black,   #...or text will be invisible
-    #'fillRule': NON_ZERO_WINDING, - these can be done later
+    'fillMode': FILL_EVEN_ODD,      #same as pdfgen.canvas
 
     'fontSize': 10,
     'fontName': _baseGFontName,
     'textAnchor':  'start' # can be start, middle, end, inherited
     }
 
-
 ####################################################################
 # math utilities.  These could probably be moved into lib
 # somewhere.
@@ -171,7 +172,6 @@ def _rotatedBoxLimits( x, y, w, h, angle):
     Y = [x[1] for x in C]
     return min(X), max(X), min(Y), max(Y), C
 
-
 class _DrawTimeResizeable:
     '''Addin class to provide the horribleness of _drawTimeResize'''
     def _drawTimeResize(self,w,h):
@@ -188,7 +188,6 @@ class _SetKeyWordArgs:
         for key, value in keywords.items():
             setattr(self, key, value)
 
-
 #################################################################
 #
 #    Helper functions for working out bounds
@@ -452,8 +451,10 @@ class Group(Shape):
         ''' return a fully expanded object'''
         from reportlab.graphics.widgetbase import Widget
         obj = Group()
+        if hasattr(self,'__label__'):
+            obj.__label__=self.__label__
         if hasattr(obj,'transform'): obj.transform = self.transform[:]
-        P = self.contents[:]    # pending nodes
+        P = self.getContents()[:]   # pending nodes
         while P:
             n = P.pop(0)
             if isinstance(n, UserNode):
@@ -506,7 +507,6 @@ class Group(Shape):
         """Convenience to help you set transforms"""
         self.transform = mmult(self.transform, scale(sx, sy))
 
-
     def skew(self, kx, ky):
         """Convenience to help you set transforms"""
         self.transform = mmult(mmult(self.transform, skewX(kx)),skewY(ky))
@@ -581,14 +581,23 @@ def _repr(self,I=None):
         return 'EmptyClipPath'
     elif isinstance(self,Shape):
         if I: _addObjImport(self,I)
-        from inspect import getargs
-        args, varargs, varkw = getargs(self.__init__.__func__.__code__)
+        from inspect import getargspec
+        args, varargs, varkw, defaults = getargspec(self.__init__)
+        if defaults:
+            kargs = args[-len(defaults):]
+            del args[-len(defaults):]
+        else:
+            kargs = []
         P = self.getProperties()
         s = self.__class__.__name__+'('
         for n in args[1:]:
             v = P[n]
             del P[n]
             s = s + '%s,' % _repr(v,I)
+        for n in kargs:
+            v = P[n]
+            del P[n]
+            s = s + '%s=%s,' % (n,_repr(v,I))
         for n,v in P.items():
             v = P[n]
             s = s + '%s=%s,' % (n, _repr(v,I))
@@ -604,10 +613,12 @@ def _renderGroupPy(G,pfx,I,i=0,indent='\t\t'):
     for n in C:
         if isinstance(n, Group):
             npfx = 'v%d' % i
-            i = i + 1
-            s = s + '%s%s=%s._nn(Group())\n' % (indent,npfx,pfx)
+            i += 1
+            l = getattr(n,'__label__','')
+            if l: l='#'+l
+            s = s + '%s%s=%s._nn(Group())%s\n' % (indent,npfx,pfx,l)
             s = s + _renderGroupPy(n,npfx,I,i,indent)
-            i = i - 1
+            i -= 1
         else:
             s = s + '%s%s.add(%s)\n' % (indent,pfx,_repr(n,I))
     return s
@@ -626,10 +637,27 @@ class Drawing(Group, Flowable):
     This has no properties except a height, width and list
     of contents."""
 
-    _saveModes=(
-            'pdf','ps','eps','gif','png','jpg','jpeg','pct',
-            'pict','tiff','tif','py','bmp','svg','tiffp','tiffl','tiff1',
-            )
+    _saveModes = {
+            'bmp',
+            'eps',
+            'gif',
+            'jpeg',
+            'jpg',
+            'pct',
+            'pdf',
+            'pict',
+            'png',
+            'ps',
+            'py',
+            'svg',
+            'tif',
+            'tiff',
+            'tiff1',
+            'tiffl',
+            'tiffp',
+            }
+
+    _bmModes = _saveModes - {'eps','pdf','ps','py','svg'}
 
     _xtraAttrMap = AttrMap(
         width = AttrMapValue(isNumber,desc="Drawing width in points."),
@@ -641,12 +669,13 @@ class Drawing(Group, Flowable):
         #AR temporary hack to track back up.
         #fontName = AttrMapValue(isStringOrNone),
         renderScale = AttrMapValue(isNumber,desc="Global scaling for rendering"),
+        initialFontName = AttrMapValue(isStringOrNone,desc="override the STATE_DEFAULTS value for fontName"),
+        initialFontSize = AttrMapValue(isNumberOrNone,desc="override the STATE_DEFAULTS value for fontSize"),
         )
 
     _attrMap = AttrMap(BASE=Group,
             formats = AttrMapValue(SequenceOf(
-                OneOf('pdf','gif','png','tif','jpg','tiff','pct','pict',
-                        'bmp','tiffp','tiffl','tiff1','eps','svg','ps','py'),
+                OneOf(*_saveModes),
                 lo=1,emptyOK=0), desc='One or more plot modes'),
             )
     _attrMap.update(_xtraAttrMap)
@@ -661,7 +690,10 @@ class Drawing(Group, Flowable):
         self.renderScale = 1.0
 
     def _renderPy(self):
-        I = {'reportlab.graphics.shapes': ['_DrawingEditorMixin','Drawing','Group']}
+        I = {
+                'reportlab.graphics.shapes': ['_DrawingEditorMixin','Drawing','Group'],
+                'reportlab.lib.colors': ['Color','CMYKColor','PCMYKColor'],
+            }
         G = _renderGroupPy(self._explode(),'self',I)
         n = 'ExplodedDrawing_' + self.__class__.__name__
         s = '#Autogenerated by ReportLab guiedit do not edit\n'
@@ -679,7 +711,8 @@ class Drawing(Group, Flowable):
         draw itself in a story.  It is specific to PDF and should not
         be used directly."""
         from reportlab.graphics import renderPDF
-        renderPDF.draw(self, self.canv, 0, 0, showBoundary=showBoundary)
+        renderPDF.draw(self, self.canv, 0, 0,
+                showBoundary=showBoundary if showBoundary is not _unset_ else getattr(self,'_showBoundary',_unset_))
 
     def wrap(self, availWidth, availHeight):
         width = self.width
@@ -762,7 +795,7 @@ class Drawing(Group, Flowable):
                 macfs.FSSpec(filename).SetCreatorType("CARO", "PDF ")
                 macostools.touched(filename)
 
-        for bmFmt in ('gif','png','tif','jpg','tiff','pct','pict', 'bmp','tiffp','tiffl','tiff1'):
+        for bmFmt in self._bmModes:
             if bmFmt in plotMode:
                 from reportlab.graphics import renderPM
                 filename = '%s.%s' % (fnroot,bmFmt)
@@ -826,13 +859,13 @@ class Drawing(Group, Flowable):
 
     def asString(self, format, verbose=None, preview=0, **kw):
         """Converts to an 8 bit string in given format."""
-        assert format in ('pdf','ps','eps','gif','png','jpg','jpeg','bmp','ppm','tiff','tif','py','pict','pct','tiffp','tiffl','tiff1'), 'Unknown file format "%s"' % format
+        assert format in self._saveModes, 'Unknown file format "%s"' % format
         from reportlab import rl_config
         #verbose = verbose is not None and (verbose,) or (getattr(self,'verbose',verbose),)[0]
         if format == 'pdf':
             from reportlab.graphics import renderPDF
             return renderPDF.drawToString(self)
-        elif format in ('gif','png','tif','tiff','jpg','pct','pict','bmp','ppm','tiffp','tiffl','tiff1'):
+        elif format in self._bmModes:
             from reportlab.graphics import renderPM
             return renderPM.drawToString(self, fmt=format,showBoundary=getattr(self,'showBorder',
                             rl_config.showBoundary),**_extraKW(self,'_renderPM_',**kw))
@@ -850,6 +883,9 @@ class Drawing(Group, Flowable):
             return renderPS.drawToString(self, showBoundary=getattr(self,'showBorder',rl_config.showBoundary))
         elif format == 'py':
             return self._renderPy()
+        elif format == 'svg':
+            from reportlab.graphics import renderSVG
+            return renderSVG.drawToString(self,showBoundary=getattr(self,'showBorder',rl_config.showBoundary),**_extraKW(self,'_renderSVG_',**kw))
 
     def resized(self,kind='fit',lpad=0,rpad=0,bpad=0,tpad=0):
         '''return a base class drawing which ensures all the contents fits'''
@@ -917,7 +953,6 @@ class LineShape(Shape):
         self.strokeOpacity = None
         self.setProperties(kw)
 
-
 class Line(LineShape):
     _attrMap = AttrMap(BASE=LineShape,
         x1 = AttrMapValue(isNumber,desc=""),
@@ -945,6 +980,7 @@ class SolidShape(LineShape):
         fillOpacity = AttrMapValue(isOpacity,desc="the level of transparency of the color, any real number between 0 and 1"),
         fillOverprint = AttrMapValue(isBoolean,desc='Turn on fill overprinting'),
         overprintMask = AttrMapValue(isBoolean,desc='overprinting for ordinary CMYK',advancedUsage=1),
+        fillMode = AttrMapValue(OneOf(FILL_EVEN_ODD,FILL_NON_ZERO)),
         )
 
     def __init__(self, kw):
@@ -954,31 +990,50 @@ class SolidShape(LineShape):
         #the above settings
         LineShape.__init__(self, kw)
 
-
 # path operator  constants
 _MOVETO, _LINETO, _CURVETO, _CLOSEPATH = list(range(4))
 _PATH_OP_ARG_COUNT = (2, 2, 6, 0)  # [moveTo, lineTo, curveTo, closePath]
 _PATH_OP_NAMES=['moveTo','lineTo','curveTo','closePath']
 
-def _renderPath(path, drawFuncs):
+def _renderPath(path,drawFuncs,countOnly=False,forceClose=False):
     """Helper function for renderers."""
     # this could be a method of Path...
     points = path.points
     i = 0
     hadClosePath = 0
     hadMoveTo = 0
+    active = not countOnly
     for op in path.operators:
+        if op == _MOVETO:
+            if forceClose:
+                if hadMoveTo and pop!=_CLOSEPATH:
+                    hadClosePath += 1
+                    if active:
+                        drawFuncs[_CLOSEPATH]()
+            hadMoveTo += 1
         nArgs = _PATH_OP_ARG_COUNT[op]
-        func = drawFuncs[op]
         j = i + nArgs
-        func(*points[i:j])
+        drawFuncs[op](*points[i:j])
         i = j
         if op == _CLOSEPATH:
-            hadClosePath = hadClosePath + 1
-        if op == _MOVETO:
-            hadMoveTo += 1
+            hadClosePath += 1
+        pop = op
+    if forceClose and hadMoveTo and pop!=_CLOSEPATH:
+        hadClosePath += 1
+        if active:
+            drawFuncs[_CLOSEPATH]()
     return hadMoveTo == hadClosePath
 
+_fillModeMap = {
+        None: None,
+        FILL_NON_ZERO: FILL_NON_ZERO,
+        'non-zero': FILL_NON_ZERO,
+        'nonzero': FILL_NON_ZERO,
+        FILL_EVEN_ODD: FILL_EVEN_ODD,
+        'even-odd': FILL_EVEN_ODD,
+        'evenodd': FILL_EVEN_ODD,
+        }
+
 class Path(SolidShape):
     """Path, made up of straight lines and bezier curves."""
 
@@ -986,9 +1041,11 @@ class Path(SolidShape):
         points = AttrMapValue(isListOfNumbers),
         operators = AttrMapValue(isListOfNumbers),
         isClipPath = AttrMapValue(isBoolean),
+        autoclose = AttrMapValue(NoneOr(OneOf('svg','pdf'))),
+        fillMode = AttrMapValue(OneOf(FILL_EVEN_ODD,FILL_NON_ZERO)),
         )
 
-    def __init__(self, points=None, operators=None, isClipPath=0, **kw):
+    def __init__(self, points=None, operators=None, isClipPath=0, autoclose=None, fillMode=None, **kw):
         SolidShape.__init__(self, kw)
         if points is None:
             points = []
@@ -998,6 +1055,7 @@ class Path(SolidShape):
         self.points = points
         self.operators = operators
         self.isClipPath = isClipPath
+        self.autoclose=autoclose
 
     def copy(self):
         new = self.__class__(self.points[:], self.operators[:])
@@ -1137,7 +1195,6 @@ class Rect(SolidShape):
     def getBounds(self):
         return (self.x, self.y, self.x + self.width, self.y + self.height)
 
-
 class Image(SolidShape):
     """Bitmap image."""
 
@@ -1294,7 +1351,7 @@ class Wedge(SolidShape):
                 a(centerx+radius1*c)
                 a(centery+yradius1*s)
         if self.annular:
-            P = Path()
+            P = Path(fillMode=getattr(self,'fillMode',None))
             P.moveTo(points[0],points[1])
             for x in xrange(2,2*n,2):
                 P.lineTo(points[x],points[x+1])
@@ -1445,6 +1502,10 @@ class UserNode(_DrawTimeResizeable):
 
         raise NotImplementedError("this method must be redefined by the user/programmer")
 
+class DirectDraw(Shape):
+    """try to draw directly on the canvas"""
+    def drawDirectly(self,canvas):
+        raise NotImplementedError("this method must be redefined by the user/programmer")
 
 def test():
     r = Rect(10,10,200,50)
@@ -1463,6 +1524,5 @@ def test():
     w('verifying...')
     r.verify()
 
-
 if __name__=='__main__':
     test()
